#include "improved_rnn.h"
#include <cstring>
#include <cmath>
#include <algorithm>
#include <adept.h>

#pragma GCC optimize("Ofast","unroll-loops","omit-frame-pointer","inline", "unsafe-math-optimizations")
#pragma GCC option("arch=native","tune=native","no-zero-upper")

using namespace nikolas;

rnn::rnn(unsigned lexicSize, unsigned hiddenSize)
    :lexicSize(lexicSize), hiddenSize(hiddenSize)
{
    // allocating internal variables once
    Wxh = new float*[hiddenSize];
    Whh = new float*[hiddenSize];
    Why = new float*[lexicSize];
    by = new float[lexicSize];
    bh = new float[hiddenSize];
    Hprev = new float[hiddenSize];
    memset(Hprev,0.0,hiddenSize*sizeof(float));
    tempHs = new float[hiddenSize];
    xs = new float[lexicSize];
    ys = new float[lexicSize];
    for(unsigned i = 0; i < hiddenSize; i++){
        Wxh[i] = new float[lexicSize];
        Whh[i] = new float[hiddenSize];
    }
    for(unsigned i = 0; i < lexicSize; i++)
        Why[i] = new float[hiddenSize];
    //std::cout << " Constructor " << std::endl;
    // allocating internal variables once
    aWxh = new adouble*[hiddenSize];
    aWhh = new adouble*[hiddenSize];
    aWhy = new adouble*[lexicSize];
    aby = new adouble[lexicSize];
    abh = new adouble[hiddenSize];
    aHprev = new adouble[hiddenSize];
    atempHs = new adouble[hiddenSize];
    axs = new adouble[lexicSize];
    ays = new adouble[lexicSize];
    for(unsigned i = 0; i < hiddenSize; i++){
        aHprev[i] = 0;
        aWxh[i] = new adouble[lexicSize];
        aWhh[i] = new adouble[hiddenSize];
    }
    for(unsigned i = 0; i < lexicSize; i++)
        aWhy[i] = new adouble[hiddenSize];
}

rnn::~rnn(){
    for(int i = hiddenSize - 1; i >=0; i--){
        delete[] Wxh[i];
        delete[] Whh[i];
    }
    for(int i = lexicSize - 1; i >= 0 ; i--)
        delete[] Why[i];
    delete[] Wxh;
    delete[] Whh;
    delete[] Why;
    delete[] by;
    delete[] bh;
    delete[] Hprev;
    delete[] tempHs;
    delete[] xs;
    delete[] ys;
    for(int i = hiddenSize - 1; i >=0; i--){
        delete[] aWxh[i];
        delete[] aWhh[i];
    }
    for(int i = lexicSize - 1; i >= 0 ; i--)
        delete[] aWhy[i];
    delete[] aWxh;
    delete[] aWhh;
    delete[] aWhy;
    delete[] aby;
    delete[] abh;
    delete[] aHprev;
    delete[] atempHs;
    delete[] axs;
    delete[] ays;
    output.clear();
}

void rnn::setParameters(std::vector<adept::adouble> &x){
    unsigned icount=0,i,j;
    //copy first hiddenSize*lexicSize to Wxh
    for(i=0;i<hiddenSize;i++)
    {
        for(j=0;j<lexicSize;j++)
        {
            aWxh[i][j]=x[icount++];
        }
    }
    //copy second hiddenSize * hiddenSize to Whh
    for(i=0;i<hiddenSize;i++)
    {
        for(j=0;j<hiddenSize;j++)
            aWhh[i][j]=x[icount++];
    }
    //copy third lexicSize * hiddenSize to Why
    for(i=0;i<lexicSize;i++)
    {
        for(j=0;j<hiddenSize;j++)
        {
            aWhy[i][j]=x[icount++];
        }
    }
    for(i=0;i<hiddenSize;i++)
        abh[i]=x[icount++];
    for(i=0;i<lexicSize;i++)
        aby[i]=x[icount++];


    memset(Hprev,0.0,hiddenSize*sizeof(float));
}

void rnn::setParameters(std::vector<double> &x)
{
    unsigned icount=0,i,j;
    //copy first hiddenSize*lexicSize to Wxh
    for(i=0;i<hiddenSize;i++)
    {
        for(j=0;j<lexicSize;j++)
        {
            Wxh[i][j]=x[icount++];
        }
    }
    //copy second hiddenSize * hiddenSize to Whh
    for(i=0;i<hiddenSize;i++)
    {
        for(j=0;j<hiddenSize;j++)
            Whh[i][j]=x[icount++];
    }
    //copy third lexicSize * hiddenSize to Why
    for(i=0;i<lexicSize;i++)
    {
        for(j=0;j<hiddenSize;j++)
        {
            Why[i][j]=x[icount++];
        }
    }
    for(i=0;i<hiddenSize;i++)
        bh[i]=x[icount++];
    for(i=0;i<lexicSize;i++)
        by[i]=x[icount++];


    memset(Hprev,0.0,hiddenSize*sizeof(float));

}

unsigned rnn::parameterSize() const
{
    return hiddenSize * lexicSize+
            hiddenSize*hiddenSize+
            lexicSize*hiddenSize+hiddenSize+lexicSize;
}

void rnn::updateaHs(){
    unsigned i,j,k;
    // tanh(Wxh*xs + Whh*hprev + bh)
    for(i = 0; i < hiddenSize; i++){
        atempHs[i] = abh[i];
        for(j = 0; j < lexicSize; j++){
            atempHs[i] += aWxh[i][j]*axs[j];
        }
        for(k = 0; k < hiddenSize; k++){
            atempHs[i] += aWhh[i][k]*aHprev[k];
        }

        atempHs[i] = tanh(atempHs[i]);
    }
    for(unsigned i = 0 ; i < hiddenSize; i++) aHprev[i] = atempHs[i];

}
void rnn::computeaYs(){
    //output[pointCounter] = 0;
    unsigned i,j;
    adouble sum = 0;

    for(i = 0; i < lexicSize; i++){

        ays[i] = aby[i];
        for(j = 0; j < hiddenSize; j++){
            ays[i] += aWhy[i][j]*aHprev[j];
        }
        ays[i] = (ays[i]*ays[i]);
        sum += ays[i];
    }

    sum = 1.0/sum;
    for(i = 0; i < lexicSize;i++){
        ays[i]*=sum;
    }
}


void rnn::updateHs(){
    unsigned i,j,k;
    // tanh(Wxh*xs + Whh*hprev + bh)
    for(i = 0; i < hiddenSize; i++){
        tempHs[i] = bh[i];
#pragma omp simd linear(j)
        for(j = 0; j < lexicSize; j++){
            tempHs[i] += Wxh[i][j]*xs[j];
        }

//#pragma omp ordered
//{
#pragma omp simd linear(k)
        for(k = 0; k < hiddenSize; k++){
            tempHs[i] += Whh[i][k]*Hprev[k];
        }

        tempHs[i] = tanh(tempHs[i]);


    }
//}
    memcpy(Hprev,tempHs,hiddenSize*sizeof(float));


}

void rnn::computeYs(){
    //output[pointCounter] = 0;
    unsigned i,j;
    float sum = 0;
    //float maxVal=0.0;
    //exp(Why*hs + by)

    for(i = 0; i < lexicSize; i++){

        ys[i] = by[i];
#pragma omp simd linear(j)
        for(j = 0; j < hiddenSize; j++){
            ys[i] += Why[i][j]*Hprev[j];
        }
        ys[i] = (ys[i]*ys[i]);
        sum += ys[i];
    }

    sum = 1/sum;
    for(i = 0; i < lexicSize;i++){
        //printf("%f ",ys[i]);
        ys[i]*=sum;
        //printf("%f ",ys[i]);
    }
    //puts("");
    //getchar();
}

std::vector<unsigned> rnn::getState(){
    return output;
}



double rnn::computeError(vector<unsigned> &in){
    //puts("Compute error");
    //printf("size %d \n",in.size());
    unsigned i;
    memset(Hprev,0.0,hiddenSize*sizeof(float));
    memset(xs,0,lexicSize*sizeof(float));
    memset(ys,0,lexicSize*sizeof(float));
    double error = 0;
    for(i = 0; i < in.size() - 1 ; i++){
        pointCounter = i;
        memset(xs,0,lexicSize*sizeof(float));
        xs[in[i]] = 1;

        updateHs();
        computeYs();
        error+=(1.0-ys[in[i+1]])*(1.0-ys[in[i+1]]);
    }

    return error;
}

double rnn::computeNaive(vector<unsigned> &in){
    unsigned i,j;

    unsigned maxIndex;
    memset(Hprev,0.0,hiddenSize*sizeof(float));
    double error = 0;
    for(i = 0; i < in.size() - 1 ; i++){
        float max = 0.0;
        pointCounter = i;
        memset(xs,0,lexicSize*sizeof(float));
        xs[in[i]] = 1;
        updateHs();
        computeYs();
        for(j=0; j < lexicSize; j++){
            if(ys[j] > max){
                max = ys[j];
                maxIndex = j;
            }
        }
        error += (in[i+1]==maxIndex)?0:1;
    }
    return error;
}

double rnn::computeFit(const vector<unsigned> &in){
    unsigned i,j;
    float max = 0.0;
    unsigned maxIndex;
    memset(Hprev,0.0,hiddenSize*sizeof(float));
    double error = 0;
    for(i = 0; i < in.size() - 1 ; i++){
        pointCounter = i;
        memset(xs,0,lexicSize*sizeof(float));
        xs[in[i]] = 1;
        updateHs();
        computeYs();

        if(max_element(ys, ys+lexicSize) - ys != in[i+1]){
            error += (0.5 + 0.5*(1-ys[in[i+1]])*(1-ys[in[i+1]]));
        }
        else{
            error += (1-ys[in[i+1]])*(1-ys[in[i+1]]);
        }//*/
        /*
        if( ( j = max_element(ys, ys+lexicSize) - ys) != in[i+1]){
            //error += (0.5 + 0.5*(1-ys[in[i+1]])*(1-ys[in[i+1]]));
            error += ys[j] + 0.5*(1-ys[in[i+1]]);
        }
        else{
            error += 0.5*(1-ys[in[i+1]]);
            for(unsigned k = 0; k < in[i+1];k++){
                error += 0.5 * (1.0/lexicSize) * (ys[k]);
            }
            for(unsigned k = in[i+1]+1; k < lexicSize;k++){
                error += 0.5 * (1.0/lexicSize) * (ys[k]);
            }
        }//*/
        /*
        error += 0.9*(1-ys[in[i+1]])*(1-ys[in[i+1]]);
        for(unsigned k = 0; k < in[i+1];k++){
            error += 0.1 * (1.0/lexicSize) * (ys[k])* (ys[k]);
        }
        for(unsigned k = in[i+1]+1; k < lexicSize;k++){
            error += 0.1 * (1.0/lexicSize) * (ys[k])* (ys[k]);
        }//*/
        /*
        error += (1.0/lexicSize) * (1-ys[in[i+1]])*(1-ys[in[i+1]]);
        for(unsigned k = 0; k < in[i+1];k++){
            error += (1.0/lexicSize) * (ys[k])* (ys[k]);
        }
        for(unsigned k = in[i+1]+1; k < lexicSize;k++){
            error += (1.0/lexicSize) * (ys[k])* (ys[k]);
        }*/
    }

    return error;
}

adouble rnn::acomputeFit(const vector<unsigned> &in){
    unsigned i,j;
    for(unsigned i = 0; i < hiddenSize;i++) aHprev[i] = 0;
    adouble error = 0;
    for(i = 0; i < in.size() - 1 ; i++){
        for(unsigned j = 0; j < lexicSize;j++) axs[j] = 0;
        axs[in[i]] = 1;
        updateaHs();
        computeaYs();

        if(max_element(ays, ays+lexicSize) - ays != in[i+1]){
            error += (0.5 + 0.5*(1-ays[in[i+1]])*(1-ays[in[i+1]]));
        }
        else{
            error += (1-ays[in[i+1]])*(1-ays[in[i+1]]);
        }//*/
        /*
        if( ( j = max_element(ays, ays+lexicSize) - ays) != in[i+1]){
            error += 0.5*ays[j] + 0.5 * (1-ays[in[i+1]]);
        }
        else{
            error += 0.5*(1-ays[in[i+1]]);
            for(unsigned k = 0; k < in[i+1];k++){
                error += 0.5 * (1.0/lexicSize) * (ays[k]);
            }
            for(unsigned k = in[i+1]+1; k < lexicSize;k++){
                error += 0.5 * (1.0/lexicSize) * (ays[k]);
            }
        }//*/
        /*
        error += 0.9*(1-ays[in[i+1]])*(1-ays[in[i+1]]);
        for(unsigned k = 0; k < in[i+1];k++){
            error += 0.1 * (1.0/lexicSize) * (ays[k])* (ays[k]);
        }
        for(unsigned k = in[i+1]+1; k < lexicSize;k++){
            error += 0.1 * (1.0/lexicSize) * (ays[k])* (ays[k]);
        }//*/
        /*
        error += (1.0/lexicSize) * (1-ays[in[i+1]])*(1-ays[in[i+1]]);
        for(unsigned k = 0; k < in[i+1];k++){
            error += (1.0/lexicSize) * (ays[k])* (ays[k]);
        }
        for(unsigned k = in[i+1]+1; k < lexicSize;k++){
            error += (1.0/lexicSize) * (ays[k])* (ays[k]);
        }*/
    }

    return error;
}


unsigned rnn::getHiddenSize() const{
    return hiddenSize;
}

unsigned rnn::getLexicSize() const{
    return lexicSize;
}

float rnn::logProb(const vector<unsigned> &in){
    unsigned i;
    float logprob = 0.0;
    memset(Hprev,0.0,hiddenSize*sizeof(float));
    for(i = 0; i < in.size() - 1 ; i++){
        pointCounter = i;
        memset(xs,0,lexicSize*sizeof(float));
        xs[in[i]] = 1;
        updateHs();
        computeYs();

        logprob += log2(ys[in[i+1]]+1);
    }
    return logprob;
}
